import json
import os

import tensorflow as tf
from tensorflow.keras.layers import Dense

import tensorflow_recommenders_addons as tfra
import tensorflow_recommenders_addons.dynamic_embedding as de
import numpy as np
import pandas as pd

from absl import app
from collections import OrderedDict
from absl import flags

from tensorflow.python.ops import io_ops
from tensorflow.python.training import checkpoint_management
from tensorflow.python.platform import gfile

flags.DEFINE_string('model_dir', "./ckpt", 'export_dir')
flags.DEFINE_string('export_dir', "./export_dir", 'export_dir')
flags.DEFINE_string('mode', "train", 'train or export')
flags.DEFINE_string('warm_path', 'ckpt1', 'warm start path')

FLAGS = flags.FLAGS
tf_config = json.loads(os.environ.get('TF_CONFIG') or '{}')
task_name = tf_config.get('task', {}).get('type', "chief")
task_idx = tf_config.get('task', {}).get('index', 0)
ps_num = len(tf_config.get('cluster', {}).get('ps', ["localhost:2223"]))


def restore_tfra_variable(ckpt_path, variable):
    mhts = variable._tables
    key_tensor_names = []
    value_tensor_names = []
    for mht in mhts:
        assert len(mht.saveable.specs) == 2
        key_tensor_names.append(mht.saveable.specs[0].name)
        value_tensor_names.append(mht.saveable.specs[1].name)

    latest_ckpt = _get_checkpoint_filename(ckpt_path)
    restore_op = io_ops.restore_v2(latest_ckpt,
                                   key_tensor_names + value_tensor_names, [""] *
                                   len(key_tensor_names + value_tensor_names),
                                   ([tf.int64] * len(key_tensor_names)) +
                                   ([tf.float32] * len(value_tensor_names)))

    key_tensor = restore_op[:len(key_tensor_names)]
    value_tensor = restore_op[len(key_tensor_names):]

    mht_restore_ops = []
    index = 0
    for mht in mhts:
        mht_restore_ops.append(
            mht.saveable.restore((key_tensor[index], value_tensor[index]), None))
        index += 1
    return mht_restore_ops


class RestoreTfraVariableHook(tf.estimator.SessionRunHook):

    def __init__(self, ckpt_path, variables):
        self._ckpt_path = ckpt_path
        self._variables = variables

    def begin(self):
        self._restore_ops = []
        for var in self._variables:
            self._restore_ops.extend(restore_tfra_variable(self._ckpt_path, var))

    def after_create_session(self, session, coord):
        session.run(self._restore_ops)



def _get_checkpoint_filename(ckpt_dir_or_file):
    if gfile.IsDirectory(ckpt_dir_or_file):
        return checkpoint_management.latest_checkpoint(ckpt_dir_or_file)
    return ckpt_dir_or_file


def input_fn():
    movie_id = np.random.randint(0, 10, (1000, 1))
    user_id = np.random.randint(0, 10, (1000, 1))
    dataset = {
        "movie_id": movie_id,
        "user_id": user_id,
        "user_rating": movie_id.astype(float) * 2.1 * user_id.astype(float) * 3 + 10
    }

    dataset = tf.data.Dataset.from_tensor_slices(dataset).repeat(100).batch(256)
    return dataset



def model_fn(features, labels, mode, params):
    embedding_size = 32
    movie_id = features["movie_id"]
    user_id = features["user_id"]
    rating = tf.reshape(features["user_rating"], (-1,))
    for k in features:
        print("k>>>", features[k].shape)

    print("mode>>>", mode)
    if mode == tf.estimator.ModeKeys.TRAIN:
        ps_list = [
            "/job:ps/replica:0/task:{}/CPU:0".format(i)
            for i in range(params["ps_num"])
        ]
        initializer = tf.keras.initializers.RandomNormal(-1, 1)
    else:
        ps_list = None#["/job:localhost/replica:0/task:0/CPU:0"] * params["ps_num"]
        initializer = tf.keras.initializers.Zeros()

    name_list = ["user_id", "movie_id"]
    feature_list = [user_id, movie_id]
    ids = tf.concat(feature_list, axis=1)  # [None, 1] concat
    print("ids>>>shape", ids.shape)
    print("ps_list", ps_list)
    embeddings = tfra.dynamic_embedding.get_variable(
        name="embeddings",
        dim=embedding_size,
        devices=ps_list,
        trainable=False,
        initializer=initializer)
    # feature_ids = tf.concat(feature_list, axis=1)
    emb_shape = tf.concat([tf.shape(ids), [embedding_size]], axis=0)

    # embs, trainable_wrapper = de.embedding_lookup_unique(embeddings, feature_ids)
    id_val, id_idx = tf.unique(tf.reshape(ids, (-1,)))
    unique_embs, trainable_wrapper = de.embedding_lookup(embeddings, id_val, return_trainable=True, name="lookup")
    #unique_embs.graph.add_to_collection(de.GraphKeys.DYNAMIC_EMBEDDING_VARIABLES, embeddings)
    flat_emb = tf.gather(unique_embs, id_idx)
    emb_lookuped = tf.reshape(flat_emb, emb_shape)

    input_emb = tf.reshape(emb_lookuped, [-1, tf.shape(emb_lookuped)[1] * tf.shape(emb_lookuped)[2]])

    d0 = Dense(256,
               activation='relu',
               kernel_initializer=tf.keras.initializers.RandomNormal(0.0, 0.1),
               bias_initializer=tf.keras.initializers.RandomNormal(0.0, 0.1))
    d1 = Dense(64,
               activation='relu',
               kernel_initializer=tf.keras.initializers.RandomNormal(0.0, 0.1),
               bias_initializer=tf.keras.initializers.RandomNormal(0.0, 0.1))
    d2 = Dense(1,
               kernel_initializer=tf.keras.initializers.RandomNormal(0.0, 0.1),
               bias_initializer=tf.keras.initializers.RandomNormal(0.0, 0.1))
    dnn = d0(input_emb)
    dnn = d1(dnn)
    dnn = d2(dnn)
    out = tf.reshape(dnn, shape=[-1])

    loss = tf.keras.losses.MeanSquaredError()(rating, out)

    # predictions = {"uid": tf.reshape(tf.concat(feature_list, 0), (-1,)), "out": flat_emb}
    predictions = {
        "uid": tf.reshape(ids, (-1,)),
        "out": flat_emb
    }

    if mode == tf.estimator.ModeKeys.TRAIN:
        optimizer = tf.compat.v1.train.AdamOptimizer(learning_rate=0.001)
        optimizer = tfra.dynamic_embedding.DynamicEmbeddingOptimizer(optimizer)
        train_op = optimizer.minimize(
            loss, global_step=tf.compat.v1.train.get_or_create_global_step())
        log_hook = tf.compat.v1.estimator.LoggingTensorHook({"emb_size": embeddings.size()}, every_n_iter=10)
        # return tf.estimator.EstimatorSpec(mode=mode, predictions=predictions)
        training_chief_hooks = None
        if params["warm_path"]:
            print("warm start", params["warm_path"])
            restore_hook = de.WarmStartHook(params["warm_path"], [embeddings])
            #restore_hook = RestoreTfraVariableHook(params["warm_path"], [embeddings])
            training_chief_hooks = [restore_hook]
        return tf.estimator.EstimatorSpec(mode=mode,
                                          predictions=predictions,
                                          loss=loss,
                                          train_op=train_op,
                                          training_hooks=[log_hook],
                                          training_chief_hooks=training_chief_hooks)

    elif mode == tf.estimator.ModeKeys.PREDICT:
        prediction_hooks = None
        if False:#params["warm_path"]:
            print("warm start", params["warm_path"])
            #restore_hook = de.WarmStartHook(params["warm_path"], [".*embeddings.*"])
            restore_hook = RestoreTfraVariableHook(params["warm_path"], [embeddings])
            prediction_hooks = [restore_hook]
        return tf.estimator.EstimatorSpec(mode, predictions=predictions, prediction_hooks=prediction_hooks)
    elif mode == tf.estimator.ModeKeys.EVAL:
        eval_metric_ops = {}
        return tf.estimator.EstimatorSpec(mode=mode, loss=loss, eval_metric_ops=eval_metric_ops)


def train(model_dir):
    params = {
        "ps_num": ps_num,
        "warm_path": FLAGS.warm_path
    }
    model_config = tf.estimator.RunConfig(log_step_count_steps=100,
                                          save_summary_steps=100,
                                          save_checkpoints_steps=100,
                                          save_checkpoints_secs=None,
                                          keep_checkpoint_max=2)
    # Save checkpoints
    estimator = tf.estimator.Estimator(model_fn=model_fn,
                                       model_dir=model_dir,
                                       params=params,
                                       config=model_config)

    # ==========  执行任务  ========== #
    if FLAGS.mode == "train":
        # early_stop_hook 是控制模型早停的控件，下面两个分别是 tf 1.x 和 tf 2.x 的写法
        # early_stop_hook = tf.contrib.estimator.stop_if_no_increase_hook(estimator, metric_name="accuracy",
        early_stop_hook = tf.estimator.experimental.stop_if_no_increase_hook(estimator, metric_name="accuracy",
                                                                             max_steps_without_increase=1000,
                                                                             min_steps=500)
        train_spec = tf.estimator.TrainSpec(input_fn=input_fn)
        eval_spec = tf.estimator.EvalSpec(input_fn=input_fn)
        tf.estimator.train_and_evaluate(estimator, train_spec, eval_spec)
        # estimator.train(train_spec)
    elif FLAGS.mode == "eval":
        estimator.evaluate(input_fn=input_fn)
    elif FLAGS.mode == "infer":
        outs = list(estimator.predict(input_fn=input_fn, predict_keys=None))
        df = pd.DataFrame(map(lambda x: x["out"], outs))
        df.index = map(lambda x: x["uid"], outs)
        df.index.name = "uid"
        df = df.reset_index().drop_duplicates(["uid"] + list(range(32))).sort_values("uid")

        df.to_csv("./result.csv", sep="\t", index=False)


def serving_input_receiver_dense_fn():
    input_spec = {
        "movie_id": tf.constant([1], tf.int64),
        "user_id": tf.constant([1], tf.int64),
        "user_rating": tf.constant([1.0], tf.float32)
    }
    return tf.estimator.export.build_raw_serving_input_receiver_fn(input_spec)


def export_for_serving(model_dir, export_dir, ps_num):
    model_config = tf.estimator.RunConfig(log_step_count_steps=100,
                                          save_summary_steps=100,
                                          save_checkpoints_steps=100,
                                          save_checkpoints_secs=None)

    estimator = tf.estimator.Estimator(model_fn=model_fn,
                                       model_dir=model_dir,
                                       params={"ps_num": ps_num},
                                       config=model_config)

    estimator.export_saved_model(export_dir, serving_input_receiver_dense_fn())


def main(argv):
    del argv

    train(FLAGS.model_dir)

    # if FLAGS.mode == "warmup":
    #     warmup_train(FLAGS.model_dir, ps_num)
    # if FLAGS.mode == "serving" and task_name == "chief" and int(task_idx) == 0:
    #     tfra.dynamic_embedding.enable_inference_mode()
    #     export_for_serving(FLAGS.model_dir, FLAGS.export_dir, ps_num)


if __name__ == "__main__":
    app.run(main)

